#include "DrawingMethod.h"

namespace Graph
{
  /**************************************************************
   ѥ饹
  **************************************************************/

  /*
    GPatternSet::operator() : ѥ˽ä

    const Coord<int>& c : 
  */
  void GPatternSet::operator()( const Coord<int>& c )
  {
    if ( ( pattern >> ( PATTERN_SIZE - index - 1 ) & 1 ) )
      draw->pset( c, col );

    index = ( index + 1 ) % PATTERN_SIZE;
  }

  /*
    GNotSet::operator() : ȿž

    const Coord<int>& c : 
  */
  void GNotSet::operator()( const Coord<int>& c )
  {
    RGB p;

    draw->point( c, p );
    draw->pset( c, ~p );
  }

  /*
    GXorSet::operator() : XOR

    const Coord<int>& c : 
  */
  void GXorSet::operator()( const Coord<int>& c )
  {
    RGB p;

    draw->point( c, p );
    p ^= col;

    draw->pset( c, p );
  }

  /*
    GBrightSet::operator() : ѹ

    const Coord<int>& c : 
  */
  void GBrightSet::operator()( const Coord<int>& c )
  {
    RGB p;
    unsigned char r, g, b;
    double d;

    draw->point( c, p );
    p.get( r, g, b );
    d = (double)r * rate;
    r = ( d > UCHAR_MAX ) ? UCHAR_MAX : (unsigned char)d;
    d = (double)g * rate;
    g = ( d > UCHAR_MAX ) ? UCHAR_MAX : (unsigned char)d;
    d = (double)b * rate;
    b = ( d > UCHAR_MAX ) ? UCHAR_MAX : (unsigned char)d;

    p.set( r, g, b );

    draw->pset( c, p );
  }

  /**************************************************************
   GLine : ʬѥ饹
  **************************************************************/

  /*
    GLine::calcSeqCode : üʬॳɤ

    const Coord<int>& c : ɸ

     : üʬॳ
  */
  int GLine::calcSeqCode( const Coord<int>& c ) const
  {
    int code = 0;
    if ( c.x < 0 ) code += LEFT;
    if ( c.x >= 0 && (unsigned int)c.x >= drawSize.x ) code += RIGHT;
    if ( c.y < 0 ) code += TOP;
    if ( c.y >= 0 && (unsigned int)c.y >= drawSize.y ) code += BOTTOM;

    return( code );
  }

  /*
    GLine::calcClipX : Ūˡˤλ()

    const Coord<int> &c0, &c1 : åԥʬɸ
    int clipX : (ޤϱ)դXɸ
    Coord<int>& c : åԥ󥰸κɸ

     : åԥ true , ʬϴԲĻ false
  */
  bool GLine::calcClipX( const Coord<int>& c0, const Coord<int>& c1, int clipX, Coord<int>& c ) const
  {
    if ( c1.x == c0.x )
      return( c0.x >= 0 && (unsigned int)c0.x < drawSize.x );

    int cy = ( c1.y - c0.y ) * ( clipX - c0.x ) / ( c1.x - c0.x ) + c0.y;

    // ꥢξFalse֤
    if ( cy < 0 ) return( false );
    if ( (unsigned int)cy >= drawSize.y ) return( false );

    c.x = clipX;
    c.y = cy;

    return( true );
  }

  /*
    GLine::calcClipY : Ūˡˤλ(岼)

    const Coord<int> &c0, &c1 : åԥʬɸ
    int clipY : (ޤϲ)դYɸ
    Coord<int>& c : åԥ󥰸κɸ

     : åԥ true , ʬϴԲĻ false
  */
  bool GLine::calcClipY( const Coord<int>& c0, const Coord<int>& c1, int clipY, Coord<int>& c ) const
  {
    if ( c1.y == c0.y )
      return( c0.y >= 0 && (unsigned int)c0.y < drawSize.y );

    int cx = ( c1.x - c0.x ) * ( clipY - c0.y ) / ( c1.y - c0.y ) + c0.x;

    // ꥢξFalse֤
    if ( cx < 0 ) return( false );
    if ( (unsigned int)cx >= drawSize.x ) return( false );

    c.x = cx;
    c.y = clipY;

    return( true );
  }

  /*
    GLine::calcClippedPoint : åԥ󥰸κɸ

    int code : üʬॳ
    const Coord<int> &c0, &c1 : ʬκɸ
    Coord<int>& c : åԥ󥰸κɸ

     : åԥ true , ʬϴԲĻ false
  */
  bool GLine::calcClippedPoint( int code, const Coord<int>& c0, const Coord<int>& c1, Coord<int>& c ) const
  {
    // ɥκü곰¦ˤ
    if ( ( code & LEFT ) != 0 )
      if ( calcClipX( c0, c1, 0, c ) )
	return( true );

    // ɥαü곰¦ˤ
    if ( ( code & RIGHT ) != 0 )
      if ( calcClipX( c0, c1, drawSize.x, c ) )
	return( true );

    // ɥξü곰¦ˤ
    if ( ( code & TOP ) != 0)
      if ( calcClipY( c0, c1, 0, c ) )
	return( true );

    // ɥβü곰¦ˤ
    if ( ( code & BOTTOM ) != 0 )
      if ( calcClipY( c0, c1, drawSize.y, c ) )
	return( true );

    return( false );  // åԥ󥰤ʤä硢ʬϴԲĻ
  }

  /*
    GLine::clipping : åԥ󥰥ᥤ롼

    Coord<int>& c0 : κɸ
    Coord<int>& c1 : κɸ

     :
      >0 ... åԥ󥰤줿
      =0 ... åԥ󥰤ɬפʤ
      <0 ... ʬϴԲĻ
  */
  int GLine::clipping( Coord<int>& c0, Coord<int>& c1 ) const
  {
    int code0, code1; // üʬॳ

    code0 = calcSeqCode( c0 );  // üʬॳɤ
    code1 = calcSeqCode( c1 );  // üʬॳɤ

    // üʬॳɤɤ0000ʤХåԥ󥰤ɬפʤ
    if ( ( code0 == 0 ) && ( code1 == 0 ) ) return( 0 );

    // üʬॳɤѤ󣰤ʤʬϴԲĻ
    if ( ( code0 & code1 ) != 0 ) return( -1 );

    Coord<int> pre0 = c0;
    Coord<int> pre1 = c1;

    // Υåԥ
    if( code0 != 0 )
      if ( ! calcClippedPoint( code0, pre0, pre1, c0 ) )
	return( -1 );

    // Υåԥ
    pre0 = c0;
    if ( code1 != 0 )
      if ( ! calcClippedPoint( code1, pre0, pre1, c1 ) )
	return( -1 );

    return( 1 );
  }

  /*
    GLine::preProcess : ʬ롼(åԥ󥰤ȵΥλ)

    Coord<int> &c0, &c1 : κɸ
    Coord<unsigned int>& d : ֤εΥ
    Coord<int>& s : 

     : True ... ɬ / False ... 
  */
  bool GLine::preProcess( Coord<int>& c0, Coord<int>& c1, Coord<unsigned int>& d, Coord<int>& s )
  {
    if ( clipping( c0, c1 ) < 0 ) return( false );

    d.x = (unsigned int)( ( c1.x > c0.x ) ? c1.x - c0.x : c0.x - c1.x );
    d.y = (unsigned int)( ( c1.y > c0.y ) ? c1.y - c0.y : c0.y - c1.y );

    s.x = ( c1.x > c0.x ) ? 1 : -1;
    s.y = ( c1.y > c0.y ) ? 1 : -1;

    return( true );
  }

  /*
    GLine::operator() : Bresenhamʬȯ르ꥺѤ饤롼

    GPixelOp& pset : Ѵؿ֥
  */
  void GLine::operator()( GPixelOp& pset )
  {
    Coord<unsigned int> d;
    Coord<int> s;
    Coord<int> c0 = coord0;
    Coord<int> c1 = coord1;
    unsigned int cnt = 0; // 

    if ( ! preProcess( c0, c1, d, s ) )
      return;

    /* 1ʲξ */
    if( d.x >= d.y ) {
      int E = -d.x;
      for( unsigned int i = 0 ; i <= d.x ; i++ ) {
	pset( c0 );
	c0.x += s.x;
	E += 2 * d.y;
	if( E >= 0 ) {
	  c0.y += s.y;
	  E -= 2 * d.x;
	}
	++cnt;
      }
    /* 1礭 */
    } else {
      int E = -d.y;
      for( unsigned int i = 0 ; i <= d.y ; i++ ) {
	pset( c0 );
	c0.y += s.y;
	E += 2 * d.x;
	if( E >= 0 ) {
	  c0.x += s.x;
	  E -= 2 * d.y;
	}
	++cnt;
      }
    }
  }

  /****************************************************************
   GClipByLine : ꤷľ¦(ޤϲ¦)褹
  ****************************************************************/

  /*
    󥹥ȥ饯

    const GPSetBase& _pset : ˻Ȥؿ֥
    const Coord<int>& c1, c2 : ľ
    bool _isUp : ľξ¦褹Ȥtrue
  */
  GClipByLine::GClipByLine( GPSetBase& _pset, const Coord<int>& c1, const Coord<int>& c2, bool _isUp )
    :GPSetBase( *( _pset.drawingArea() ) ), pset( &_pset ), isUp( _isUp )
  {
    // ľľȤʤν
    if ( isVert = ( c1.x == c2.x ) ) {
      b = c1.x;
      return;
    }

    a = (double)( c1.y - c2.y ) / (double)( c1.x - c2.x );
    b = (double)( c1.y ) - a * (double)( c1.x );
  }

  /*
    

    const Coord<int>& c : 
  */
  void GClipByLine::operator()( const Coord<int>& c )
  {
    if ( pset == 0 ) return;

    if ( isVert ) {
      if ( ( isUp && (double)( c.x ) >= b ) || ( ( ! isUp ) && (double)( c.x ) <= b ) )
	(*pset)( c );
    } else if ( ( isUp && (double)( c.y ) >= (double)( c.x ) * a + b ) ||
		( ( ! isUp ) && (double)( c.y ) <= (double)( c.x ) * a + b ) ) {
      (*pset)( c );
    }
  }

  /*
    GClipByLineκ(إѴؿ)

    ľˤ륯åԥ󥰤

    GPSetBase& pset : Ѵؿ֥
    const Coord<int>& c1, c2 : ϡλ

     : GClipByLine֥
  */
  GClipByLine createGClipByLine( GPSetBase& pset, const Coord<int>& sc, const Coord<int>& ec )
  {
    bool isUp = ec.x - sc.x <= 0;
    GClipByLine gClipByLine( pset, sc, ec, isUp );

    return( gClipByLine );
  }

  /****************************************************************
   GClipByAngle : ꤷ褹
  ****************************************************************/

  /*
    󥹥ȥ饯

    const GPSetBase& _pset : ˻Ȥؿ֥
    const Coord<int>& o : 濴κɸ
    int sa, ea : ϡλ
    bool _isOut : ϰϳ褹ʤtrue
  */
  GClipByAngle::GClipByAngle( GPSetBase& _pset, const Coord<int>& o, int sa, int ea, bool _isOut )
    : GPSetBase( *( _pset.drawingArea() ) ), pset( &_pset ), origin( o ), isOut( _isOut )
  {
    while ( sa < 0 ) sa += tMax;
    while ( ea < 0 ) ea += tMax;
    sa %= tMax;
    ea %= tMax;

    sArg = (unsigned int)sa;
    eArg = (unsigned int)ea;
  }

  /*
    󥹥ȥ饯

    const GPSetBase& _pset : ˻Ȥؿ֥
    const Coord<int>& o : 濴κɸ
    const Coord<int>& c1, c2 : ϰϤκɸ
    bool _isOut : ϰϳ褹ʤtrue
  */
  GClipByAngle::GClipByAngle( GPSetBase& _pset, const Coord<int>& o, const Coord<int>& c1, const Coord<int>& c2, bool _isOut )
    : GPSetBase( *( _pset.drawingArea() ) ), pset( &_pset ), origin( o ), isOut( _isOut )
  {
    double a1 = getAngleFromCoord( c1 );
    double a2 = getAngleFromCoord( c2 );
    if ( c1.x > c2.x ) {
      sArg = (unsigned int)a2;
      eArg = (unsigned int)a1;
    } else {
      sArg = (unsigned int)a1;
      eArg = (unsigned int)a2;
    }
  }

  /*
    ꤵ줿ɸ顢Xγ(065536)

    const Coord<int>& c : ɸ

     : ᤿
  */
  double GClipByAngle::getAngleFromCoord( const Coord<int>& c ) const
  {
    Coord<int> rel( c ); // 濴кɸ
    rel -= origin;

    double angle = ( rel.x != 0 ) ?
      atan( (double)( rel.y ) / (double)( rel.x ) ) :
      ( rel.y >= 0 ) ? M_PI / 2 : -M_PI / 2;

    if ( rel.x < 0 ) angle += M_PI;
    if ( angle < 0 ) angle += M_PI * 2;
    angle = angle * tMax / ( M_PI * 2 );

    return( angle );
  }

  /*
    

    const Coord<int>& c : 
  */
  void GClipByAngle::operator()( const Coord<int>& c )
  {
    double angle = getAngleFromCoord( c );

    if ( sArg > eArg ) {
      if ( ( ( ! isOut ) && ( angle < sArg && angle > eArg ) ) ||
	   ( (   isOut ) && ( angle >= sArg || angle <= eArg ) ) )
	(*pset)( c );
    } else {
      if ( ( ( ! isOut ) && ( angle < sArg || angle > eArg ) ) ||
	   ( (   isOut ) && ( angle >= sArg && angle <= eArg ) ) )
	(*pset)( c );
    }
  }

  /*
    GClipByAngleκ(إѴؿ)

    ٤ˤ륯åԥ󥰤

    GPSetBase& pset : Ѵؿ֥
    const Coord<int>& o
    int sAngle, eAngle : ϡλ
    const Coord<int>& c1, c2 : ϰϤ򼨤κɸ

     : GClipByLine֥
  */
  GClipByAngle createGClipByAngle( GPSetBase& pset, const Coord<int>& o, int sAngle, int eAngle )
  {
    GClipByAngle gClipByAngle( pset, o, sAngle, eAngle );

    return( gClipByAngle );
  }
  GClipByAngle createGClipByAngle( GPSetBase& pset, const Coord<int>& o, const Coord<int>& c1, const Coord<int>& c2 )
  {
    GClipByAngle gClipByAngle( pset, o, c1, c2 );

    return( gClipByAngle );
  }

  /**************************************************************
   GEllipse : ʱѥ饹
  **************************************************************/

  /*
    󥹥ȥ饯

    гȤĹܤʱߤ褹
    (ʿľξϤľ¤Ȥ뿿ߤ褹)

    const Coord<int>& c1,c2 : гξü
    const Coord<unsigned int>& s : 礭
    bool _fillMode : ɤĤ֤⡼
  */
  GEllipse::GEllipse( const Coord<int>& c1, const Coord<int>& c2, const Coord<unsigned int>& s, bool _fillMode )
    : drawSize( s )
  {
    if ( ! getParamByRectangle( c1, c2 ) ) {
      origin = Coord<int>( roundHalfUp( (double)( c1.x + c2.x ) / 2 ), roundHalfUp( (double)( c1.y + c2.y ) / 2 ) );
      r = (unsigned int)( origin.distance( c1 ) + 0.5 );
      k = Coord<unsigned int>( 1, 1 );
    }
    fillMode = _fillMode;
  }

  /*
    гȤĹܤʱߤΥѥ᡼
    (ʱߤΥѥ᡼̵¤¸ߤ뤿ᡢʱߤ濴гüؤкɸ(u,v)Ф
      k = Coord( v, u ), r = uv( u + v ) ɽͤȤ)

    const Coord<int>& c1, c2 : гκɸ

     : ĹʬȤʤfalse֤
  */
  bool GEllipse::getParamByRectangle( const Coord<int>& c1, const Coord<int>& c2 )
  {
    if ( c1.x == c2.x || c1.y == c2.y ) return( false );

    Coord<double> o( (double)( c1.x + c2.x ) / 2, (double)( c1.y + c2.y ) / 2 );
    Coord<double> p( (double)( max( c1.x, c2.x ) ) - o.x, (double)( max( c1.y, c2.y ) ) - o.y );

    k = Coord<unsigned int>( (unsigned int)( p.y + 0.5 ), (unsigned int)( p.x + 0.5 ) );
    r = (unsigned int)( sqrt( ( p.x + p.y ) * p.x * p.y ) + 0.5 );
    origin = Coord<int>( roundHalfUp( o.x ), roundHalfUp( o.y ) );

    return( true );
  }

  /*
    Xγt(065536)ˤκɸ

    unsigned int t : 

     : ᤿ɸ
  */
  Coord<int> GEllipse::getCoordFromAngle( unsigned int t ) const
  {
    const unsigned int tMax = 65536;
    const unsigned int t2 = tMax >> 2;
    const unsigned int t3 = tMax >> 1;
    const unsigned int t4 = ( tMax >> 2 ) * 3;

    t %= tMax;

    if ( t == t2 ) // Y
      return( Coord<int>( origin.x, (int)( (double)r / ::sqrt( k.y ) ) + origin.y ) );
    if ( t == t4 ) // Y
      return( Coord<int>( origin.x, (int)( (double)r / ::sqrt( k.y ) ) * -1 + origin.y ) );

    double powM = ::pow( tan( 2 * M_PI * (double)t / (double)tMax ), 2 ); // t뷹
    Coord<int> c;

    c.x = (int)( (double)r / ::sqrt( (double)( k.x ) + (double)( k.y ) * powM ) );
    c.y = (int)( (double)r * sqrt( powM ) / ::sqrt( (double)( k.x ) + (double)( k.y ) * powM ) );

    if ( t >= t2 && t < t4 ) c.x *= -1;
    if ( t >= t3 ) c.y *= -1;

    c.x += origin.x;
    c.y += origin.y;

    return( c );
  }

  bool GEllipse::checkYRange( int y ) const
  {
    if ( origin.y + y < 0 ) return( false ); // ˥ꥢξ¦
    if ( origin.y - y < 0 ) return( true );  // ꥢξ¦ȸ򺹤

    return( (unsigned int)( origin.y - y ) < drawSize.y );
  }

  /*
    ʱؿ

    a( x - origin.x )^2 + b( y - origin.y )^2 = r^2褹

    GPixelOp& pset : Ѵؿ֥
  */
  void GEllipse::operator()( GPixelOp& pset )
  {
    if ( r == 0 ) return; // Ⱦ¤0ʤ褷ʤ

    Coord<int> c0( (int)( (double)r / sqrt( (double)( k.x ) ) + 0.5 ), 0 ); // 賫Ϻɸ
    double d = sqrt( (double)( k.x ) ) * (double)r;
    int F = (int)( -2.0 * d + 0.5 ) +     k.x + 2 * k.y;
    int H = (int)( -4.0 * d + 0.5 ) + 2 * k.x     + k.y;
    bool drawn = false; // оݤYɸǤѤߤ(ɤĤ֤⡼ɤξͭ)

    while ( c0.x >= 0 ) {
      if ( checkYRange( c0.y ) ) {
	// ɤĤ֤⡼
	if ( fillMode && ( ! drawn ) ) {
	  int x = origin.x - c0.x;
	  if ( x < 0 ) x = 0;
	  int maxX = origin.x + c0.x;
	  if ( maxX < 0 ) maxX = 0;
	  if ( (unsigned int)maxX >= drawSize.x ) maxX = drawSize.x - 1;
	  for ( ; x <= maxX ; ++x ) {
	    if ( origin.y - c0.y >= 0 ) pset( Coord<int>( x, origin.y - c0.y ) );
	    if ( origin.y + c0.y >= 0 )
	      if ( (unsigned int)( origin.y + c0.y ) < drawSize.y && c0.y != 0 )
		pset( Coord<int>( x, origin.y + c0.y ) );
	  }
	// ̾α߸
	} else {
	  pset( Coord<int>( origin.x + c0.x, origin.y + c0.y ) );
	  if ( c0.x != 0 ) pset( Coord<int>( origin.x - c0.x, origin.y + c0.y ) );
	  if ( c0.y != 0 ) {
	    pset( Coord<int>( origin.x + c0.x, origin.y - c0.y ) );
	    if ( c0.x != 0 ) pset( Coord<int>( origin.x - c0.x, origin.y - c0.y ) );
	  }
	}
      }
      if ( F >= 0 ) {
	--( c0.x );
	F -= 4 * k.x * c0.x;
	H -= 4 * k.x * c0.x - 2 * k.x;
	drawn = true;
      }
      if ( H < 0 ) {
	++( c0.y );
	F += 4 * k.y * c0.y + 2 * k.y;
	H += 4 * k.y * c0.y;
	drawn = false;
      }
    }
  }

  /**************************************************************
   GCircle : ѥ饹
  **************************************************************/

  /*
    󥹥ȥ饯

    ߼λ濴Ⱦ¤ꤹ

    const Coord<int>& c1,c2,c3 : ߼̤뻰
    const Coord<unsigned int>& s : 礭
    bool _fillMode : ɤĤ֤⡼
  */
  GCircle::GCircle( const Coord<int>& c1, const Coord<int>& c2, const Coord<int>& c3, const Coord<unsigned int>& s, bool _fillMode )
  {
    origin() = calcOrigin( c1, c2, c3, r );
    drawSize = s;
    fillMode() = _fillMode;
  }

  /*
    󥹥ȥ饯

    ľ¤ξüȤƱߤꤹ

    const Coord<int>& c1,c2 : ߼̤
    const Coord<unsigned int>& s : 礭
    bool _fillMode : ɤĤ֤⡼
  */
  GCircle::GCircle( const Coord<int>& c1, const Coord<int>& c2, const Coord<unsigned int>& s, bool _fillMode )
  {
    origin() = Coord<int>( ( c1.x + c2.x + 1 ) / 2, ( c1.y + c2.y + 1 ) / 2 ); // "+ 1"ϻͼθΤ
    r = (unsigned int)( origin().distance( c1 ) );
    drawSize = s;
    fillMode() = _fillMode;
  }

  /*
    ֤εΥǺΤΤФƤȵΥȾʬ
    (Ʊʬˤä濴Ⱦ¤)

    const Coord<int>& c1, c2, c3 : κɸ
    unsigned int& d : ξüȤεΥ

     : Υˤʤ
  */
  Coord<int> GCircle::calcMidPoint( const Coord<int>& c1, const Coord<int>& c2, const Coord<int>& c3, unsigned int& d )
  {
    double d12 = c1.distance( c2 );
    double d23 = c2.distance( c3 );
    double d31 = c3.distance( c1 );

    // "+ 1"ϻͼθΤ
    if ( d12 > d23 && d12 > d31 ) {
      d = (unsigned int)( ( d12 + 1 ) / 2 );
      return( Coord<int>( ( c1.x + c2.x + 1 ) / 2, ( c1.y + c2.y + 1 ) / 2 ) );
    } else if ( d23 > d31 && d23 > d12 ) {
      d = (unsigned int)( ( d23 + 1 ) / 2 );
      return( Coord<int>( ( c2.x + c3.x + 1 ) / 2, ( c2.y + c3.y + 1 ) / 2 ) );
    } else {
      d = (unsigned int)( ( d31 + 1 ) / 2 );
      return( Coord<int>( ( c3.x + c1.x + 1 ) / 2, ( c3.y + c1.y + 1 ) / 2 ) );
    }
  }

  /*
    ʬοľʬФȷ

    const Coord<int>& c1, c2 : ʬξü
    Coord<double>& m : ʬ
    double& a : ľʬη

     : ʬYɸפfalse֤
  */
  bool GCircle::calcPerpendicularBisector( const Coord<int>& c1, const Coord<int>& c2, Coord<double>& m, double& a )
  {
    if ( c1.y == c2.y ) return( false );

    m = Coord<double>( (double)( c1.x + c2.x ) / 2, (double)( c1.y + c2.y ) / 2 );
    a = (double)( c1.x - c2.x ) / (double)( c2.y - c1.y );

    return( true );
  }

  /*
    ѷγȾ¤

    const Coord<int>& c1, c2, c3 : ѷκɸ
    unsigned int& _r : Ⱦ

     : 濴ɸ (ʤ c1c3ǤǤʬ֤)
  */
  Coord<int> GCircle::calcOrigin( const Coord<int>& c1, const Coord<int>& c2, const Coord<int>& c3, unsigned int& _r )
  {
    // ֤ȷ
    Coord<double> m1, m2;
    double a1, a2;

    if ( c1.y == c2.y && c1.y == c3.y )
      return( calcMidPoint( c1, c2, c3, _r ) ); // ʿˤϵʤ

    if ( c1.y == c2.y ) {
      calcPerpendicularBisector( c2, c3, m1, a1 );
      calcPerpendicularBisector( c3, c1, m2, a2 );
    } else if ( c2.y == c3.y ) {
      calcPerpendicularBisector( c1, c2, m1, a1 );
      calcPerpendicularBisector( c3, c1, m2, a2 );
    } else {
      calcPerpendicularBisector( c1, c2, m1, a1 );
      calcPerpendicularBisector( c2, c3, m2, a2 );
    }

    if ( a1 == a2 )
      return( calcMidPoint( c1, c2, c3, _r ) ); // ľ¤Ǥϵʤ

    double x = ( a1 * m1.x - a2 * m2.x - m1.y + m2.y );
    double y = ( a1 * a2 * ( m1.x - m2.x ) - a2 * m1.y + a1 * m2.y );
    _r = (unsigned int)( sqrt( ::pow( ( ( a1 - a2 ) * (double)( c1.x ) - x ) / ( a1 - a2 ), 2 ) +
			       ::pow( ( ( a1 - a2 ) * (double)( c1.y ) - y ) / ( a1 - a2 ), 2 ) ) + 0.5 );

    x /= a1 - a2;
    y /= a1 - a2;

    Coord<int> ans( roundHalfUp( x ), roundHalfUp( y ) );

    return( ans );
  }

  /*
    濴ȱ߼ΰꤷƱ߸̤褹

    const Coord<int>& o : 濴ɸ
    const Coord<int>& p : ߼κɸ
    GPixelOp& pset : Ѵؿ֥
  */
  void GCircle::operator()( const Coord<int>& o, const Coord<int>& p, GPixelOp& pset )
  {
    r = (unsigned int)( sqrt( ::pow( ( o.x - p.x ), 2 ) + ::pow( ( o.y - p.y ), 2 ) ) + 0.5 );
    origin() =  o;

    (*this)( pset );
  }

  /*****************************
   NotDraw : ȿžѥ饹
  *****************************/

  /*
    ޥ֤ϤƤȿž

    const Coord<int>& q : ޥݥ󥿤κɸ
  */
  void NotDraw::notDraw( const Coord<int>& q )
  {
    if ( inProcess() ) {
      notDraw(); // Ƥ򥯥ꥢ
    } else {
      init( q );
    }
    set( q );
    notDraw();
  }

  /*
    ޥ֤ϤƤ̾

    const Coord<int>& q : ޥݥ󥿤κɸ
  */
  void NotDraw::draw( const Coord<int>& q )
  {
    if ( ! inProcess() ) return;

    clear(); // Ƥ򥯥ꥢ
    set( q );
    draw();
  }

  /***************************************
   NotDraw2Pass : 2ѥȿžѥ饹
  ***************************************/

  /*
    黻Ҥ¿

    const NotDraw2Pass& preDraw : ԡ
  */
  NotDraw2Pass& NotDraw2Pass::operator=( const NotDraw2Pass& preDraw )
  {
    derive( preDraw );

    return( *this );
  }

  /*
    Υꥢ
  */
  void NotDraw2Pass::clear()
  {
    if ( inProcess() ) notDraw();
    inProc = false;
  }

  /***************************************
   NotLine : ȿžʬѥ饹
  ***************************************/

  void NotLine::draw( GPSetBase& pset )
  {
    line.drawSize = pset.size();
    line( p0, p1, pset );
  }

  /***************************************
   NotCircle : ȿž߸ѥ饹
  ***************************************/

  void NotCircle::draw( GPSetBase& pset )
  {
    circle.size() = pset.size();
    circle( p0, p1, pset );
  }

  /****************************************
    NotCircleWithDia : ľդ߸
  ****************************************/

  /*
    ľդ߸ȿž
  */
  void NotCircleWithDia::notDraw()
  {
    line.drawSize = notSet->size();
    line( p0, p1, *notSet );
    draw( *notSet );
  }

  /*
    ľդ߸̾

    GPSetBase& pset : Ѵؿ֥
  */
  void NotCircleWithDia::draw( GPSetBase& pset )
  {
    circle = GCircle( p0, p1, pset.size(), circle.fillMode() );
    circle( pset );
  }

  /*********************
    NotBox : 
  *********************/

  /*
    

    GPixelOp& pset : Ѵؿ֥
  */
  void NotBox::draw( GPSetBase& pset )
  {
    Coord<int> lu = p0;
    Coord<int> rd = p1;

    if ( lu.x > rd.x ) swap( lu.x, rd.x );
    if ( lu.y > rd.y ) swap( lu.y, rd.y );

    line.drawSize = pset.size();
    line( lu, Coord<int>( rd.x, lu.y ), pset );
    if ( lu.y != rd.y )
      line( Coord<int>( lu.x, rd.y ), rd, pset );

    if ( rd.y - lu.y > 1 ) {
      line( Coord<int>( lu.x, lu.y + 1 ), Coord<int>( lu.x, rd.y - 1 ), pset );
      if ( lu.x != rd.x )
	line( Coord<int>( rd.x, lu.y + 1 ), Coord<int>( rd.x, rd.y - 1 ), pset );
    }
  }

  /*************************
    NotEllipse : ʱ
  *************************/

  /*
    ʱ̾

    GPixelOp& pset : Ѵؿ֥
  */
  void NotEllipse::draw( GPSetBase& pset )
  {
    GEllipse ellipse( p0, p1, pset.size() );
    ellipse( pset );
  }

  /*************************************************
    NotCircleWithTriangle : ѷȳܱߤ
  *************************************************/

  /*
    ѷȳܱߤȿž
  */
  void NotCircleWithTriangle::notDraw()
  {
    line.drawSize = notSet->size();
    line( p0, p1, *notSet );
    line( p1, p2, *notSet );
    line( p2, p0, *notSet );
    draw( *notSet );
  }

  /*
    ̾(ߤΤߤ)

    GPSetBase& pset : Ѵؿ֥
  */
  void NotCircleWithTriangle::draw( GPSetBase& pset )
  {
    circle = GCircle( p0, p1, p2, pset.size(), circle.fillMode() );
    circle( pset );
  }

  /*************************************************
    NotArcWithTriangle : ѷȳܱ(߸)
  *************************************************/

  /*
    ѷȳܱ(߸)ȿž
  */
  void NotArcWithTriangle::notDraw()
  {
    line.drawSize = notSet->size();
    line( p0, p1, *notSet );
    line( p1, p2, *notSet );
    line( p2, p0, *notSet );
    draw( *notSet );
  }

  /*
    ̾(߸̤Τߤ)

    GPSetBase& pset : Ѵؿ֥
  */
  void NotArcWithTriangle::draw( GPSetBase& pset )
  {
    circle = GCircle( p0, p1, p2, pset.size(), circle.fillMode() );
    GClipByAngle gClip( pset, circle.origin(), p0, p1 );
    if ( p0.x == p1.x ) {
      gClip.isOut = p0.x < p2.x;
    } else {
      double gradient = (double)( p0.y - p1.y ) / (double)( p0.x - p1.x );
      gClip.isOut = ( gradient * (double)( p2.x - p0.x ) + (double)( p0.y ) ) > (double)( p2.y );
    }
    //gClip = circle.createGClipByLine( pset, p0, p1 );
    //gClip.isUp = gClip.gradient() * (double)( p2.x ) + gClip.intercept() < (double)( p2.y );
    circle( gClip );
  }

  /**************************************************************
   DrawingArea_SelectionEvent : DrawingAreaѥ٥
  **************************************************************/

  /*
    󥹥ȥ饯

    DrawingArea& widget : оݤDrawingArea
  */
  DrawingArea_SelectionEvent::DrawingArea_SelectionEvent( NotDraw& d )
    : GtkCallback( *( d.drawingArea() ) ), draw( &d )
  {
    setCallback( "drawing_area", "button-press-event", G_CALLBACK(buttonPressEvent) );
    setCallback( "drawing_area", "motion_notify_event", G_CALLBACK(motionNotifyEvent) );
  }

  /*
    ޥ٥Ⱦμ

    GdkEventMotion *event :GdkEventExpose֥
    Coord<int>& c : ޥΰ
    GdkModifierType& state : ޥξ
  */
  void DrawingArea_SelectionEvent::getEventInfo( GdkEventMotion *event, Coord<int>& c, GdkModifierType& state )
  {
    if ( event->is_hint ) {
      gdk_window_get_pointer( event->window, &( c.x ), &( c.y ), &state );
    } else {
      c.x = (int)( event->x );
      c.y = (int)( event->y );
      state = (GdkModifierType)( event->state );
    }
  }

  /*
    ޥưΥХåؿ

    GtkWidget *widget : widget()֥
    GdkEventMotion* *event : GdkEventMotion٥
    gpointer gp : DrawingArea_SelectionEvent֥(ʻȥݥ)
  */
  gint DrawingArea_SelectionEvent::motionNotifyEvent( GtkWidget *widget, GdkEventMotion *event, gpointer gp )
  {
    DrawingArea_SelectionEvent* daEvt = (DrawingArea_SelectionEvent*) gp;

    Coord<int> c;
    GdkModifierType state;
    getEventInfo( event, c, state );

    return( daEvt->motionNotifyAction( c, state ) );
  }

  /***********************************************************************
   DrawingArea_2PassSelectionEvent : DrawingAreaѥ٥(2ѥ)
  ***********************************************************************/

  /*
    ܥ󥯥åΥХåؿ

    GdkEventButton *event : GdkEventButton٥
  */
  gint DrawingArea_2PassSelectionEvent::buttonPressAction( GdkEventButton *event )
  {
    Coord<int> c( (int)( event->x ), (int)( event->y ) );
    if ( event->button == 1 ) {
      if ( draw->inProcess() )
	draw->draw( draw->drawingArea()->pixbufCoord( c ) );
      else
	draw->notDraw( draw->drawingArea()->pixbufCoord( c ) );
    } else if ( event->button == 3 ) {
      draw->clear();
    }

    draw->drawingArea()->reflesh();

    return( FALSE );
  }

  /*
    ޥưΥХåؿ

    Coord<int>& c : ޥΰ
    GdkModifierType& state : ޥξ
  */
  gint DrawingArea_2PassSelectionEvent::motionNotifyAction( Coord<int>& c, GdkModifierType& state )
  {
    if ( draw->inProcess() )
      draw->notDraw( draw->drawingArea()->pixbufCoord( c ) );

    draw->drawingArea()->reflesh();

    return( FALSE );
  }

  /*********************************************************************************
   DrawingArea_MultiPassSelectionEvent : DrawingAreaѥ٥(ʣΥѥ)
  *********************************************************************************/

  /*
    ܥ󥯥åΥХåؿ

    GdkEventButton *event : GdkEventButton٥
  */
  /*
  gint DrawingArea_MultiPassSelectionEvent::buttonPressAction( GdkEventButton *event )
  {
    Coord<int> c( (int)( event->x ), (int)( event->y ) );
    if ( event->button == 1 ) {
      if ( draw->inProcess() )
	draw->draw( drawingArea->pixbufCoord( c ), *pset );
      else
	draw->draw( drawingArea->pixbufCoord( c ) );
    } else if ( event->button == 3 ) {
      draw->clear();
    }

    drawingArea->reflesh();

    return( FALSE );
  }
  */

  /*
    ޥưΥХåؿ

    Coord<int>& c : ޥΰ
    GdkModifierType& state : ޥξ
  */
  /*
  gint DrawingArea_MultiPassSelectionEvent::motionNotifyAction( Coord<int>& c, GdkModifierType& state )
  {
    if ( draw->inProcess() )
      draw->draw( drawingArea->pixbufCoord( c ) );

    drawingArea->reflesh();

    return( FALSE );
  }
  */

  /**********************************************************************************
   DrawingArea_TriPointSelectionEvent : 򤷤Ʊ߸̤褹뤿Υ٥
  **********************************************************************************/

  /*
    ܥ󥯥åΥХåؿ

    GdkEventButton *event : GdkEventButton٥
  */
  gint DrawingArea_TriPointSelectionEvent::buttonPressAction( GdkEventButton *event )
  {
    Coord<int> c( (int)( event->x ), (int)( event->y ) );
    if ( event->button == 1 ) {
      if ( draw == draw2 ) {
	if ( draw->inProcess() ) {
	  draw->draw( draw->drawingArea()->pixbufCoord( c ) );
	  draw = draw1;
	} else {
	  draw->notDraw( draw->drawingArea()->pixbufCoord( c ) );
	}
      } else {
	if ( draw->inProcess() ) {
	  draw->clear();
	  delete draw2;
	  draw2 = new NotArcWithTriangle( *draw1 );
	  draw = draw2;
	}
	draw->notDraw( draw->drawingArea()->pixbufCoord( c ) );
      }
    } else if ( event->button == 3 ) {
      draw->clear();
      draw = draw1;
    }

    draw->drawingArea()->reflesh();

    return( FALSE );
  }

  /*
    ޥưΥХåؿ

    Coord<int>& c : ޥΰ
    GdkModifierType& state : ޥξ
  */
  gint DrawingArea_TriPointSelectionEvent::motionNotifyAction( Coord<int>& c, GdkModifierType& state )
  {
    if ( draw->inProcess() )
      draw->notDraw( draw->drawingArea()->pixbufCoord( c ) );

    draw->drawingArea()->reflesh();

    return( FALSE );
  }

}


