#ifndef DRAWING_METHOD_H
#define DRAWING_METHOD_H

#include "pixbuf.h"

using namespace GtkPP;

namespace Graph
{
  /**************************************************************
   إѴؿ
  **************************************************************/

  /*
    doubleΥǡ򾮿ʲͼθintȤ֤
  */
  inline int roundHalfUp( double d )
  {
    if ( d > 0 )
      d += 0.5;
    else
      d -= 0.5;

    return( (int)d );
  }

  template<class T> void swap( T& t1, T& t2 )
  {
    T t = t1;
    t1 = t2;
    t2 = t;
  }

  /**********************************************
    GPixelOp : ԥѴ()饹
  **********************************************/
  class GPixelOp
  {
  public:
    virtual void operator()( const Coord<int>& c ) = 0;
    virtual ~GPixelOp() {}
  };

  /**************************************************************
   DrawingAreaФѥ饹
  **************************************************************/

  /*
    GPSetBase : ()饹
  */
  class GPSetBase : public GPixelOp
  {
  protected:

    DrawingArea* draw;

  public:

    GPSetBase( DrawingArea& d ) : draw( &d ) {}

    Coord<unsigned int> size() const { return( Coord<unsigned int>( draw->pixbufWidth(), draw->pixbufHeight() ) ); }
    DrawingArea* drawingArea() const { return( draw ); }
    GtkWidget* widget() const { return( draw->widget() ); }
  };

  /*
    GPSet : 襯饹
  */
  class GPSet : public GPSetBase
  {
  public:

    /* ѿ */
    RGB col; // 迧

    // 󥹥ȥ饯
    GPSet( DrawingArea& d, const RGB& c ) : GPSetBase( d ), col( c ) {}

    void operator()( const Coord<int>& c ) { draw->pset( c, col ); } // 
  };

  /*
    GPatternSet : ѥ˽ä襯饹
  */
  class GPatternSet : public GPSet
  {
    static const unsigned int PATTERN_SIZE = CHAR_BIT;

    char index; // ѥθ߰

  public:

    /* ѿ */
    unsigned char pattern; // ѥ

    // 󥹥ȥ饯
    GPatternSet( DrawingArea& d, const RGB& c, unsigned char _pattern = 0xFF )
      : GPSet( d, c ), index( 0 ), pattern( _pattern ) {}

    void operator()( const Coord<int>& c );
  };

  /*
    GNotSet : ȿž襯饹
  */
  class GNotSet : public GPSetBase
  {
  public:

    // 󥹥ȥ饯
    GNotSet( DrawingArea& d ) : GPSetBase( d ) {}

    void operator()( const Coord<int>& c ); // ȿž
  };

  /*
    GXorSet : XOR襯饹
  */
  class GXorSet : public GPSet
  {
  public:

    // 󥹥ȥ饯
    GXorSet( DrawingArea& d, const RGB& c ) : GPSet( d, c ) {}

    void operator()( const Coord<int>& c ); // XOR
  };

  /*
    GBrightSet : ѹ
  */
  class GBrightSet : public GPSetBase
  {
  public:

    double rate;

    // 󥹥ȥ饯
    GBrightSet( DrawingArea& d, double r ) : GPSetBase( d ), rate( r ) {}

    void operator()( const Coord<int>& c ); // ѹ
  };

  /*
    GRandomSet : ʿ襯饹
  */
  class GRandomSet : public GPSet
  {
  public:

    // 󥹥ȥ饯
    GRandomSet( DrawingArea& d )
      : GPSet( d, RGB( (int)( (double)rand() / ( (double)RAND_MAX + 1 ) * 0xffffff ) ) ) {}

    void operator()( const Coord<int>& c ) { draw->pset( c, col ); } // ʿ
  };

  /**************************************************************
   GDraw : ޷Ѵ쥯饹
  **************************************************************/
  class GDraw
  {
  public:

    virtual void operator()( GPixelOp& pset ) = 0;
  };

  /**************************************************************
   GLine : ʬѥ饹
  **************************************************************/
  class GLine : public GDraw
  {
    /* üʬॳ */
    enum Edge {
      LEFT = 1,
      RIGHT = 2,
      TOP = 4,
      BOTTOM = 8,
    };

  public:

    /* ѿ */
    Coord<int> coord0; // κɸ
    Coord<int> coord1; // κɸ
    Coord<unsigned int> drawSize; // 礭

  private:

    // üʬॳɤ
    int calcSeqCode( const Coord<int>& c ) const;

    // åԥ󥰥ꥢʬθ
    bool calcClipX( const Coord<int>& c0, const Coord<int>& c1, int clipX, Coord<int>& c ) const;
    bool calcClipY( const Coord<int>& c0, const Coord<int>& c1, int clipY, Coord<int>& c ) const;
    bool calcClippedPoint( int code, const Coord<int>& c0, const Coord<int>& c1, Coord<int>& c ) const;

    // Ѵؿ
    bool preProcess( Coord<int>& c0, Coord<int>& c1, Coord<unsigned int>& d, Coord<int>& s );

  public:

    // 󥹥ȥ饯
    GLine( int x0 = 0, int y0 = 0, int x1 = 0, int y1 = 0, unsigned int sx = 0, unsigned int sy = 0 )
      : coord0( Coord<int>( x0, y0 ) ), coord1( Coord<int>( x1, y1 ) ), drawSize( Coord<unsigned int>( sx, sy ) ) {}
    GLine( const Coord<int>& c0, const Coord<int>& c1, const Coord<unsigned int>& s )
      : coord0( c0 ), coord1( c1 ), drawSize( s ) {}

    // åԥ󥰽
    int clipping( Coord<int>& c0, Coord<int>& c1 ) const;

    // ʬؿ
    void operator()( GPixelOp& pset );
    void operator()( const Coord<int>& c0, const Coord<int>& c1, GPixelOp& pset )
      { coord0 = c0; coord1 = c1; (*this)( pset ); }
  };

  /****************************************************************
   GClipByLine : ꤷľ¦(ޤϲ¦)褹
  ****************************************************************/
  class GClipByLine : public GPSetBase
  {
    GPSetBase* pset; // δؿ֥

    double a;    // 
    double b;    // 
    bool isVert; // ľϿľ ?

  public:

    bool isUp; // ľξ¦褹ʤtrue

    double gradient() const { return( a ); }  // ֤
    double intercept() const { return( b ); } // Ҥ֤

    // 󥹥ȥ饯
    GClipByLine( GPSetBase& _pset, const Coord<int>& c1, const Coord<int>& c2, bool _isUp );

    void operator()( const Coord<int>& c ); // 
  };

  /****************************************************************
   GClipByAngle : ꤷ褹
  ****************************************************************/
  class GClipByAngle : public GPSetBase
  {
    static const unsigned int tMax = 65536; // ٤κ(=2)

    GPSetBase* pset; // δؿ֥

    Coord<int> origin; // 濴
    unsigned int sArg; // ϳ
    unsigned int eArg; // λ

    double getAngleFromCoord( const Coord<int>& c ) const; // ɸ٤

  public:

    bool isOut; // ϰϳ褹ʤtrue

    // 󥹥ȥ饯
    GClipByAngle( GPSetBase& _pset, const Coord<int>& o, int sa, int ea, bool _isOut = false );
    GClipByAngle( GPSetBase& _pset, const Coord<int>& o, const Coord<int>& c1, const Coord<int>& c2, bool _isOut = false );

    void operator()( const Coord<int>& c ); // 
  };

  /**************************************************************
   GEllipse : ʱѥ饹
  **************************************************************/
  class GEllipse : public GDraw
  {
  public:

    /* ѿ */
    Coord<int> origin;            // 濴ɸ
    unsigned int r;               // Ⱦ
    Coord<unsigned int> k;        // 
    Coord<unsigned int> drawSize; // 礭
    bool fillMode;                // ɤĤ֤⡼

  private:

    template<class T> static T max( const T& t1, const T& t2 )
      { return( ( t1 > t2 ) ? t1 : t2 ); }

    bool checkYRange( int y ) const;

    // гȤĹܤʱߤΥѥ᡼
    bool getParamByRectangle( const Coord<int>& c1, const Coord<int>& c2 );

  public:

    // 󥹥ȥ饯
    GEllipse( int ox = 0, int oy = 0, unsigned int _r = 0, unsigned int sx = 0, unsigned int sy = 0, bool _fillMode = false )
      : origin( Coord<int>( ox, oy ) ), r( _r ), k( Coord<unsigned int>( 1, 1 ) ), drawSize( Coord<unsigned int>( sx, sy ) ), fillMode( _fillMode ) {}
    GEllipse( const Coord<int>& o, unsigned int _r, const Coord<unsigned int>& _k, const Coord<unsigned int>& s, bool _fillMode = false )
      : origin( o ), r( _r ), k( _k ), drawSize( s ), fillMode( _fillMode ) {}
    GEllipse( const Coord<int>& c1, const Coord<int>& c2, const Coord<unsigned int>& s, bool _fillMode = false );

    // 濴ɸΥå
    void setOrigin( int ox, int oy ) { origin = Coord<int>( ox, oy ); }
    // Υå
    void setCoef( unsigned int kx, unsigned int ky ) { k = Coord<unsigned int>( kx, ky ); }

    // Xγt(065536)ˤκɸ
    Coord<int> getCoordFromAngle( unsigned int t ) const;

    // ʱؿ
    void operator()( GPixelOp& pset );
    void operator()( const Coord<int>& o, unsigned int _r, const Coord<unsigned int>& _k, GPixelOp& pset )
      { origin = o; r = _r; k = _k; (*this)( pset ); }
  };

  /**************************************************************
   GCircle : ѥ饹
  **************************************************************/
  class GCircle : private GEllipse
  {
    /* static function */
    // Ʊʬˤä濴Ⱦ¤
    static Coord<int> calcMidPoint( const Coord<int>& c1, const Coord<int>& c2, const Coord<int>& c3, unsigned int& d );
    // ʬοľʬФȷ
    static bool calcPerpendicularBisector( const Coord<int>& c1, const Coord<int>& c2, Coord<double>& m, double& a );
    // ѷγȾ¤
    static Coord<int> calcOrigin( const Coord<int>& c1, const Coord<int>& c2, const Coord<int>& c3, unsigned int& _r );

  public:

    // 󥹥ȥ饯
    GCircle( int ox = 0, int oy = 0, unsigned int _r = 0, unsigned int sx = 0, unsigned int sy = 0, bool _fillMode = false )
      : GEllipse( ox, oy, _r, sx, sy, _fillMode ) {}
    GCircle( const Coord<int>& o, unsigned int _r, const Coord<unsigned int>& s, bool _fillMode = false )
      : GEllipse( o, _r, Coord<unsigned int>( 1, 1 ), s, _fillMode ) {}
    GCircle( const Coord<int>& c1, const Coord<int>& c2, const Coord<int>& c3, const Coord<unsigned int>& s, bool _fillMode = false );
    GCircle( const Coord<int>& c1, const Coord<int>& c2, const Coord<unsigned int>& s, bool _fillMode = false );

    // 濴ɸ
    Coord<int>& origin() { return( GEllipse::origin ); }
    // Ⱦ
    unsigned int& radius() { return( GEllipse::r ); }
    // 
    Coord<unsigned int>& size() { return( GEllipse::drawSize ); }
    // ɤĤ֤⡼
    bool& fillMode() { return( GEllipse::fillMode ); }

    // Xγt(065536)ˤκɸ
    Coord<int> getCoordFromAngle( unsigned int t ) const
      { return( GEllipse::getCoordFromAngle( t ) ); }

    // ʱؿ
    void operator()( GPixelOp& pset ) { GEllipse::operator()( pset ); }
    void operator()( const Coord<int>& o, unsigned int _r, GPixelOp& pset )
      { origin() = o; radius() = _r; (*this)( pset ); }
    void operator()( const Coord<int>& o, const Coord<int>& p, GPixelOp& pset );
  };

  // GClipByAngle֤إѴؿ
  GClipByAngle createGClipByAngle( GPSetBase& pset, const Coord<int>& o, const Coord<int>& c1, const Coord<int>& c2 ); // ɸ
  GClipByAngle createGClipByAngle( GPSetBase& pset, const Coord<int>& o, int sAngle, int eAngle );                     // ٻ

  // GClipByLine֤إѴؿ
  GClipByLine createGClipByLine( GPSetBase& pset, const Coord<int>& c1, const Coord<int>& c2 ); // ɸ
  inline GClipByLine createGClipByLine( const GEllipse& ellipse, GPSetBase& pset, unsigned int sAngle, unsigned int eAngle )    // ٻ
    { return( createGClipByLine( pset, ellipse.getCoordFromAngle( sAngle ), ellipse.getCoordFromAngle( eAngle ) ) ); }
  inline GClipByLine createGClipByLine( const GCircle& circle, GPSetBase& pset, unsigned int sAngle, unsigned int eAngle )    // ٻ
    { return( createGClipByLine( pset, circle.getCoordFromAngle( sAngle ), circle.getCoordFromAngle( eAngle ) ) ); }

  /****************************
   NotDraw : ȿžѥ饹
  ****************************/

  /*
    쥯饹
  */
  class NotDraw
  {
  protected:

    GNotSet* notSet;
    GPSetBase* pset;

  public:

    // 󥹥ȥ饯
    NotDraw( GPSetBase& _pset ) : notSet( 0 )
      { bind( _pset ); }
    // ǥȥ饯
    virtual ~NotDraw() { delete notSet; }

    // 椫
    virtual bool inProcess() const = 0;
    // ΰ֤
    DrawingArea* drawingArea() const { return( pset->drawingArea() ); }
    // ؿ֥ȤΥХ
    void bind( GPSetBase& _pset )
      { pset = &_pset; delete notSet; notSet = new GNotSet( *( pset->drawingArea() ) ); }
    void bind( NotDraw& draw ) { bind( *( draw.pset ) ); }

    // 賫ϻν
    virtual void init( const Coord<int>& q ) = 0;
    // ǤΥѥ᡼Υå
    virtual void set( const Coord<int>& q ) = 0;
    // 轪λθ
    virtual void clear() = 0;

    // 롼
    void notDraw( const Coord<int>& q ); // ޥ֤ϤƤȿž
    void draw( const Coord<int>& q );    // ޥ֤ϤƤ̾

    virtual void notDraw() { draw( *notSet ); } // ȿž롼
    virtual void draw() { draw( *pset ); }     // ̾롼
    virtual void draw( GPSetBase& _pset ) = 0; // 롼
  };

  /*
    2ѥѴ쥯饹
  */
  class NotDraw2Pass : public NotDraw
  {
  protected:

    bool inProc;              // 椫

    Coord<int> p0;
    Coord<int> p1;

  public:

    // 󥹥ȥ饯
    NotDraw2Pass( GPSetBase& pset ) : NotDraw( pset ), inProc( false ) {}
    // ԡ󥹥ȥ饯
    // ԡ硢򼨤ե饰falseˤ
    NotDraw2Pass( const NotDraw2Pass& preDraw )
      : NotDraw( *( preDraw.pset ) ) { derive( preDraw ); }
    // ǥȥ饯
    ~NotDraw2Pass() { clear(); }

    // 椫
    bool inProcess() const { return( inProc ); }

    // 賫ϻν
    void init( const Coord<int>& q ) { p0 = q; inProc = true; }
    // ǤΥѥ᡼Υå
    void set( const Coord<int>& q ) { p1 = q; }
    // 轪λθ
    void clear();

    // NotDraw2Passѥ᡼Ѿ(򼨤ե饰falseˤ)
    void derive( const NotDraw2Pass& draw ) { p0 = draw.p0; p1 = draw.p1; inProc = false; }

    // 黻Ҥ¿
    // ԡΤ p0,p1ΤߤǡѴؿ֥Ȥϥԡʤ
    NotDraw2Pass& operator=( const NotDraw2Pass& preDraw );
  };

  /*
    ʬ
  */
  class NotLine : public NotDraw2Pass
  {
    GLine line;

  public:

    // 󥹥ȥ饯
    NotLine( GPSetBase& pset ) : NotDraw2Pass( pset ) {}

    void draw( GPSetBase& pset );
  };

  /*
    ߸
  */
  class NotCircle : public NotDraw2Pass
  {
  protected:

    GCircle circle;

  public:

    // 󥹥ȥ饯
    NotCircle( GPSetBase& pset, bool _fillMode = false ) : NotDraw2Pass( pset )
      { circle.fillMode() = _fillMode; }

    bool& fillMode() { return( circle.fillMode() ); }

    void draw( GPSetBase& pset );
  };

  /*
    ľդ߸
  */
  class NotCircleWithDia : public NotCircle
  {
  protected:

    GLine line;

  public:

    // 󥹥ȥ饯
    NotCircleWithDia( GPSetBase& pset, bool fillMode = false )
      : NotCircle( pset, fillMode ) {}

    void notDraw();
    void draw( GPSetBase& pset );
  };

  /*
    
  */
  class NotBox : public NotDraw2Pass
  {
    GLine line;

  public:

    bool fillMode;

    // 󥹥ȥ饯
    NotBox( GPSetBase& pset, bool _fillMode = false )
      : NotDraw2Pass( pset ), fillMode( _fillMode ) {}

    void draw( GPSetBase& pset );
  };

  /*
    ʱ
  */
  class NotEllipse : public NotDraw2Pass
  {
  public:

    bool fillMode;

    // 󥹥ȥ饯
    NotEllipse( GPSetBase& pset, bool _fillMode = false )
      : NotDraw2Pass( pset ), fillMode( _fillMode ) {}

    void draw( GPSetBase& pset );
  };

  /*
    ѷȳܱߤ
  */
  class NotCircleWithTriangle : public NotCircleWithDia
  {
    Coord<int> p2;

  public:

    NotCircleWithTriangle( GPSetBase& pset, bool fillMode = false )
      : NotCircleWithDia( pset, fillMode ) {}
    NotCircleWithTriangle( const NotCircleWithDia& preDraw )
      : NotCircleWithDia( preDraw ) {}

    // 賫ϻν
    void init( const Coord<int>& q ) { p2 = q; inProc = true; }
    // ǤΥѥ᡼Υå
    void set( const Coord<int>& q ) { p2 = q; }

    void notDraw();               // ȿž롼
    void draw( GPSetBase& pset ); // 롼
  };

  /*
    ѷȳܱ(߸)
  */
  class NotArcWithTriangle : public NotCircleWithDia
  {
    Coord<int> p2;
    //GClipByLine gClip;
    //GClipByAngle gClip;

  public:

    NotArcWithTriangle( GPSetBase& pset, bool fillMode = false )
      : NotCircleWithDia( pset, fillMode ) {}
    NotArcWithTriangle( const NotCircleWithDia& preDraw )
      : NotCircleWithDia( preDraw ) {}

    // 賫ϻν
    void init( const Coord<int>& q ) { p2 = q; inProc = true; }
    // ǤΥѥ᡼Υå
    void set( const Coord<int>& q ) { p2 = q; }

    void notDraw();               // ȿž롼
    void draw( GPSetBase& pset ); // ̾롼
  };

  /************************************************************
   DrawingArea_SelectionEvent : DrawingAreaѥ٥
  ************************************************************/
  class DrawingArea_SelectionEvent : public GtkPP::GtkCallback
  {
    static void getEventInfo( GdkEventMotion *event, Coord<int>& c, GdkModifierType& state );
    static gint buttonPressEvent( GtkWidget *widget, GdkEventButton *event, gpointer gp )
      { return( ( (DrawingArea_SelectionEvent*) gp )->buttonPressAction( event ) ); }
    static gint motionNotifyEvent( GtkWidget *widget, GdkEventMotion *event, gpointer gp );

  protected:

    NotDraw* draw;

    virtual gint buttonPressAction( GdkEventButton *event ) { return( FALSE ); }
    virtual gint motionNotifyAction( Coord<int>& c, GdkModifierType& state ) { return( FALSE ); }

  public:

    DrawingArea_SelectionEvent( NotDraw& d );

  };

  /***********************************************************************
   DrawingArea_2PassSelectionEvent : DrawingAreaѥ٥(2ѥ)
  ***********************************************************************/
  class DrawingArea_2PassSelectionEvent : public DrawingArea_SelectionEvent
  {
    virtual gint buttonPressAction( GdkEventButton *event );
    virtual gint motionNotifyAction( Coord<int>& c, GdkModifierType& state );

  public:

    DrawingArea_2PassSelectionEvent( NotDraw& d )
      : DrawingArea_SelectionEvent( d ) {}
  };

  /*********************************************************************************
   DrawingArea_MultiPassSelectionEvent : DrawingAreaѥ٥(ʣΥѥ)
  *********************************************************************************/
  /*
  class DrawingArea_MultiPassSelectionEvent : public DrawingArea_SelectionEvent
  {
    NotDraw* draw;

    virtual gint buttonPressAction( GdkEventButton *event );
    virtual gint motionNotifyAction( Coord<int>& c, GdkModifierType& state );

  public:

    DrawingArea_MultiPassSelectionEvent( DrawingArea& widget, NotDraw& _draw, GPSetBase& _pset )
      : DrawingArea_SelectionEvent( widget, _pset ), draw( &_draw ) { draw->bind( widget ); }
  };
  */

  /**********************************************************************************
   DrawingArea_TriPointSelectionEvent : 򤷤Ʊ߸̤褹뤿Υ٥
  **********************************************************************************/
  class DrawingArea_TriPointSelectionEvent : public DrawingArea_SelectionEvent
  {
    NotCircleWithDia* draw1;
    //NotCircleWithTriangle draw2;
    NotArcWithTriangle* draw2;

    gint buttonPressAction( GdkEventButton *event );
    gint motionNotifyAction( Coord<int>& c, GdkModifierType& state );

  public:

    // 󥹥ȥ饯
    DrawingArea_TriPointSelectionEvent( NotCircleWithDia& d )
      : DrawingArea_SelectionEvent( d ), draw1( &d ), draw2( 0 ) {}

    ~DrawingArea_TriPointSelectionEvent() { delete draw2; }
  };

}

#endif
