#ifndef GTKPLUSPLUS_H
#define GTKPLUSPLUS_H

#include <gtk/gtk.h>
#include <string>
#include <set>
#include <map>
#include <iostream>

#include "gadget.h"

namespace GtkPP
{
  using namespace std;
  using Gadget::trim;
  using Gadget::isort;

  /****************
    إѴؿ
  ****************/

  // ܥ󥦥åȤκ
  GtkWidget* newButton( const string& stockID, GtkSignalFunc clickedFunc, void* userData );
  string convFromUTF8( const gchar* c );

  /**************************************************
    GtkLessFunction : ƥѴؿ֥
  **************************************************/
  struct GtkLessFunc : public binary_function<string,string,bool>
  {
    virtual bool operator()( const string& s1, const string& s2 ) const { return( s1 < s2 ); }
    void qsort( vector<string>& vecS, unsigned int start, unsigned int end );  // Quick Sort
    void isort( vector<string>& vecS, unsigned int start, unsigned int end );  // Insertion Sort
  };

  /************************************
    GtkUTF8 : UTF8Ѵѥ֥
  ************************************/
  class GtkUTF8
  {
    gchar* gc;

  public:

    // 󥹥ȥ饯
    GtkUTF8( const string& s );
    GtkUTF8( const char* c );

    // ǥȥ饯
    ~GtkUTF8() { g_free( gc ); }

    // UTF8Ѵʸ֤
    const gchar* const value() { return( gc ); }
  };

  /*******************************************************
    Dialog : ɥ뤿Υ饹
  *******************************************************/
  class Dialog
  {
    GtkWidget* dialog; // GtkDialog

  public:

    // 󥹥ȥ饯
    Dialog( GtkWindow* root, const string& title, const string& message,
     GtkDialogFlags flags = (GtkDialogFlags)( GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT ) );

    ~Dialog() { gtk_widget_destroy( dialog ); } // ǥȥ饯

    // ܥɲ
    void addButton( const string& text, int responseId )
     { gtk_dialog_add_button( GTK_DIALOG(dialog), GtkUTF8( text ).value(), responseId ); }
    void addButton( const map<string,int>& m );  // ɲ

    // VBOXꥢؤΥåɲ
    void packStart( GtkWidget *child, bool expand = true, bool fill = true, unsigned int padding = 0 )
      { gtk_box_pack_start( GTK_BOX(GTK_DIALOG(dialog)->vbox), child, expand, fill, padding ); }
    void packEnd( GtkWidget *child, bool expand = true, bool fill = true, unsigned int padding = 0 )
      { gtk_box_pack_end( GTK_BOX(GTK_DIALOG(dialog)->vbox), child, expand, fill, padding ); }

    // ɽ(åܥФresponseId֤)
    int show() const
     { return( gtk_dialog_run( GTK_DIALOG( dialog ) ) ); }

  };

  /*************************************************
    GtkBaseWidget : åȴ쥯饹
  *************************************************/
  class GtkBaseWidget
  {
    GtkWidget* baseWidget;  // ƥå

    // ٥Ͽmap
    //   ٥ȤФ륳Хåؿ³Ѥ롣
    //   ʤФ̾դ֥ȤؤΥݥ󥿤ȤȤϿƤ
    map<string,GObject*> mapWidget;

  protected:

    // ٥ϿmapؤΥåϿ
    bool addWidget( const string& ident, GObject* gObject );
    bool removeWidget( const string& ident );

    // ƥåȤϿ
    void createBase( GtkWidget* widget ) { baseWidget = widget; }

    GObject* getWidget( const string& ident )
     { return( ( mapWidget.find( ident ) != mapWidget.end() ) ? mapWidget[ident] : 0 ); }

  public:

    // 󥹥ȥ饯
    GtkBaseWidget() : baseWidget( 0 ) {}

    // ۥǥȥ饯
    virtual ~GtkBaseWidget() {}

    // åȤΥХ
    //   ¾ΥåȤ饪֥ȤФ륳Хåؿؤ³Ǥ褦ˤ롣
    void bind( const string& classIdent, const GtkBaseWidget& widget );

    // Хåؿ³֥å֥å
    bool connect( const string& ident, const string& event, GCallback func, void* userData, unsigned int* id = 0 );
    bool disconnect( const string& ident, unsigned int id );

    void sigBlock( const string& ident, unsigned int id );
    void sigUnblock( const string& ident, unsigned int id );

    // ƥåȤ֤
    GtkWidget* widget() const { return( baseWidget ); }

    // åȤɽ
    void show() const { if ( widget() != 0 ) gtk_widget_show_all( widget() ); }

    // Ͽ줿ʤΰϤ
    void printWidget() const;
  };

  /********************************************************
    GtkCallback :Хåؿ֥ȴ쥯饹
  ********************************************************/
  class GtkCallback
  {
    struct EventName {
      string ident;
      string event;
      EventName( const string& id, const string& ev )
	: ident( id ), event( ev ) {}
      bool operator<( const EventName& e ) const
      {
	if ( ident == e.ident )
	  return( event < e.event );
	else
	  return( ident < e.ident );
      }
    };

    typedef map<EventName,GCallback> mapCallback;
    typedef map<EventName,unsigned int> mapEventId;
    typedef mapCallback::const_iterator const_iterator;

    GtkBaseWidget* widget;
    mapCallback callback;
    mapEventId eventId;
    bool connectFlag;

    static void noEvent() {} // ⤷ʤ٥

  public:

    GtkCallback( GtkBaseWidget& _widget )
      : widget( &_widget ), connectFlag( false ) {}

    virtual ~GtkCallback() {}

    bool isConnected() const { return( connectFlag ); }
    bool setCallback( const string& ident, const string& event, GCallback func = noEvent );
    bool hasCallback( const string& ident, const string& event ) const
      { return( callback.find( EventName( ident, event ) ) != callback.end() ); }

    unsigned int connect();
    unsigned int disconnect();
    void block();
    void unblock();

    void printEventName() const;
  };

  /****************************************************************
    Menu : ˥塼СΤҤȤĤΥ˥塼뤿ι¤
  ****************************************************************/
  class Menu : public GtkBaseWidget
  {
    GtkWidget* menu;   // GtkMenu

  public:

    // 󥹥ȥ饯
    Menu( const string& title );

    // ֥˥塼ɲ
    bool addItem( const string& itemName, GCallback callback, gpointer data );
  };

  /*********************************************
    MenuBar : ˥塼С뤿Υ饹
  *********************************************/
  class MenuBar : public GtkBaseWidget
  {
    GtkWidget* menuBar; // GtkMenuBar

  public:

    MenuBar() { createBase( gtk_menu_bar_new() ); } // 󥹥ȥ饯

    // ˥塼ɲ
    void addMenu( const Menu& menu ) { gtk_menu_shell_append( GTK_MENU_SHELL( widget() ), menu.widget() ); }
  };

  /*************************************************
    GtkButtonBox : ܥܥå
  *************************************************/
  class GtkButtonBox : public GtkBaseWidget
  {
    map<string,GtkWidget*> buttonList;

    int index( const string& ident );

  public:
    // 󥹥ȥ饯
    GtkButtonBox( GtkOrientation orientation );

    // ܥϿ
    bool addButton( const string& itemName, const string& stockID, GCallback callback, gpointer data );
    bool removeButton( const string& ident );
    bool insertButton( const string& ident, const string& dest, const string& stockID, GCallback callback, gpointer data );

    // ܥɽɽ
    void showButton( const string& itemName )
      { if ( buttonList.find( itemName ) != buttonList.end() ) gtk_widget_show( buttonList[itemName] ); }
    void hideButton( const string& itemName )
      { if ( buttonList.find( itemName ) != buttonList.end() ) gtk_widget_hide( buttonList[itemName] ); }
  };

  /*************************************************
    GtkEntryWithButton : ƥȥܥåդܥ

    "entry" ... ƥѥȥ
    "button" ... ܥ
  *************************************************/
  class GtkEntryWithButton : public GtkBaseWidget
  {
    GtkWidget* mainBox;        // ƥʥå
    map<string,GtkWidget*> widgets;  // Ͽmap

    // åȤκ
    void createWidget( const string& stockID, GCallback clickedFunc, void* userData );

  public:

    // 󥹥ȥ饯
    GtkEntryWithButton( const string& stockID, GCallback func = 0, void* userData = 0 )
     { createWidget( stockID, func, userData ); }

    // ǥȥ饯
    virtual ~GtkEntryWithButton() { gtk_widget_destroy( mainBox ); }

    string getText() { return( convFromUTF8( gtk_entry_get_text( GTK_ENTRY( widgets["entry"] ) ) ) ); }
    void setText( string s ) { gtk_entry_set_text( GTK_ENTRY( widgets["entry"] ), GtkUTF8( s ).value() ); }
  };

  /***************************************************************
    GtkListStoreClass : GtkListStoreѥ饹

    "model" ...  GtkTreeViewI/F(GtkTreeModel)
  ***************************************************************/
  class GtkListStoreClass : public GtkBaseWidget
  {
    GtkListStore* store;         // GtkTreeViewѥ󥿡ե(No event)
    GtkTreeModel* model;         // GtkTreeViewѥ󥿡ե

    GtkLessFunc* lessFunc;     // ƥѴؿ֥
    bool needToDeleteLessFunc; // lessFuncΤɬפ뤫

    // åȤκ
    void createWidget();

  protected:

    typedef string (*ItemFunc)( GtkListStoreClass* list, GtkTreeIter* iter, gboolean* hasIter );

    // ѽؿ
    set<string> funcAllItem( ItemFunc pf );  // ƥ
    unsigned int funcGivenItem( ItemFunc pf, const set<string>& setS );  // ꤷƥ

    // iterؤƥ
    static string getItem( GtkListStoreClass* list, GtkTreeIter* iter, gboolean* hasIter = 0 );
    static string removeItem( GtkListStoreClass* list, GtkTreeIter* iter, gboolean* hasIter = 0 );

    // ƥ¤ؤ(lessFuncѹ˻)
    void sortItem();

    // ƥΥѴؿ
    void qsort( vector<string>& vecS, unsigned int start, unsigned int end ) { lessFunc->qsort( vecS, start, end ); }
    void isort( vector<string>& vecS, unsigned int start, unsigned int end ) { lessFunc->isort( vecS, start, end ); }

    GtkListStore* getStore() const { return( store ); }
    GtkTreeModel* getModel() const { return( model ); }
    void unrefStore() { g_object_unref( G_OBJECT(store) ); }

  public:

    typedef GtkLessFunc Less;  // less_function

    // 󥹥ȥ饯
    GtkListStoreClass();
    GtkListStoreClass( GtkLessFunc* _lessFunc );
    GtkListStoreClass( const GtkListStoreClass& gtkListStoreClass );

    // ǥȥ饯
    virtual ~GtkListStoreClass();

    // ƥ
    set<string> getAllItem() { return( funcAllItem( getItem ) ); }

    // ƥɲä
    bool addItem( const string& item );
    unsigned int addItem( const set<string>& setS );

    // ƥ
    bool removeItem( const string& s );
    unsigned int removeItem( const set<string>& setS ) { return( funcGivenItem( removeItem, setS ) ); }
    set<string> removeAllItem() { return( funcAllItem( removeItem ) ); }

    // ƥ¾GtkListStoreClassذư
    set<string> moveAllItem( GtkListStoreClass* dest );

    // ƥ(ƥ֤ʤ)
    void clearItem() { gtk_list_store_clear( store ); }

    // lessFuncΥåȡꥻå(ꥻåȤˤäƥǥեȤѤ)
    void setLessFunc( GtkLessFunc* _lessFunc );
    void resetLessFunc();
  };

  /***************************************************************
    GtkComboBoxClass : GtkComboBoxѥ饹

    "model" ...  GtkTreeViewI/F(GtkTreeModel)
  ***************************************************************/
  class GtkComboBoxClass : public GtkListStoreClass
  {
    GtkWidget* combo;  // ꥹɽѥå

    void createWidget();

  public:

    // 󥹥ȥ饯
    GtkComboBoxClass();
    GtkComboBoxClass( GtkLessFunc* _lessFunc );
    GtkComboBoxClass( const GtkComboBoxClass& gtkComboBoxClass );
  };

  /***************************************************************
    GtkSingleList : ñ쥢ƥɽѥꥹȥܥå

    "list" ... ꥹɽѥå(GtkTreeView)
    "model" ...  GtkTreeViewI/F(GtkTreeModel)
    "selection" ...  ƥѥ֥(GtkTreeSelection)
    "cell" ... ѥ顦֥(GtkCellRenderer)
    "column" ... 󥪥֥(GtkTreeViewColumn)
  ***************************************************************/
  /*
  class GtkSingleList : public GtkBaseWidget
  {
    typedef string (*ItemFunc)( GtkSingleList* list, GtkTreeIter* iter, gboolean* hasIter );

    string dataName;
    GtkWidget* list;             // ꥹɽѥå
    GtkListStore* store;         // GtkTreeViewѥ󥿡ե(No event)
    GtkTreeModel* model;         // GtkTreeViewѥ󥿡ե

    GtkLessFunc* lessFunc;     // ƥѴؿ֥
    bool needToDeleteLessFunc; // lessFuncΤɬפ뤫

    // åȤκ
    void createWidget( GtkSelectionMode mode );

    // ѽؿ
    set<string> funcAllItem( ItemFunc pf );  // ƥ
    set<string> funcSelectedItem( ItemFunc pf );  // 򥢥ƥ
    unsigned int funcGivenItem( ItemFunc pf, const set<string>& setS );  // ꤷƥ

    // iterؤƥ
    static string getItem( GtkSingleList* list, GtkTreeIter* iter, gboolean* hasIter = 0 );
    static string removeItem( GtkSingleList* list, GtkTreeIter* iter, gboolean* hasIter = 0 );

    // ƥ¤ؤ(lessFuncѹ˻)
    void sortItem();

    // ƥΥѴؿ
    void qsort( vector<string>& vecS, unsigned int start, unsigned int end ) { lessFunc->qsort( vecS, start, end ); }
    void isort( vector<string>& vecS, unsigned int start, unsigned int end ) { lessFunc->isort( vecS, start, end ); }

  public:

    typedef GtkLessFunc Less;  // less_function

    // 󥹥ȥ饯
    GtkSingleList( const string& _dataName, GtkSelectionMode mode );
    GtkSingleList( const string& _dataName, GtkSelectionMode mode, Less* _lessFunc );
    GtkSingleList( const GtkSingleList& gtkSingleList );

    // ǥȥ饯
    virtual ~GtkSingleList();

    // 򥢥ƥࡦƥ
    set<string> getSelectedItem() { return( funcSelectedItem( getItem ) ); }
    set<string> getAllItem() { return( funcAllItem( getItem ) ); }

    // ƥɲä
    bool addItem( const string& item );
    unsigned int addItem( const set<string>& setS );

    // ƥ
    bool removeItem( const string& s );
    unsigned int removeItem( const set<string>& setS ) { return( funcGivenItem( removeItem, setS ) ); }
    set<string> removeSelectedItem() { return( funcSelectedItem( removeItem ) ); }
    set<string> removeAllItem() { return( funcAllItem( removeItem ) ); }

    // ƥ¾GtkSingleListذư
    set<string> moveSelectedItem( GtkSingleList* dest );
    set<string> moveAllItem( GtkSingleList* dest );

    // ƥ(ƥ֤ʤ)
    void clearItem() { gtk_list_store_clear( store ); }

    // lessFuncΥåȡꥻå(ꥻåȤˤäƥǥեȤѤ)
    void setLessFunc( Less* _lessFunc );
    void resetLessFunc();

    // ⡼ɤμ
    GtkSelectionMode getMode() const { return( gtk_tree_selection_get_mode( gtk_tree_view_get_selection( GTK_TREE_VIEW(list) ) ) ); }
    void setMode( GtkSelectionMode mode ) { gtk_tree_selection_set_mode( gtk_tree_view_get_selection( GTK_TREE_VIEW(list) ), mode ); }

    // ̾ѹ
    void setTitle( string newName );
  };
  */
  class GtkSingleList : public GtkListStoreClass
  {
    string dataName;
    GtkWidget* list;             // ꥹɽѥå

    // åȤκ
    void createWidget( GtkSelectionMode mode );

    // ѽؿ
    set<string> funcSelectedItem( ItemFunc pf );  // 򥢥ƥ

  public:

    // 󥹥ȥ饯
    GtkSingleList( const string& _dataName, GtkSelectionMode mode );
    GtkSingleList( const string& _dataName, GtkSelectionMode mode, Less* _lessFunc );
    GtkSingleList( const GtkSingleList& gtkSingleList );

    // ǥȥ饯
    virtual ~GtkSingleList();

    // 򥢥ƥ
    set<string> getSelectedItem() { return( funcSelectedItem( getItem ) ); }

    // ƥ
    set<string> removeSelectedItem() { return( funcSelectedItem( removeItem ) ); }

    // ƥ¾GtkSingleListذư
    set<string> moveSelectedItem( GtkSingleList* dest );

    // ⡼ɤμ
    GtkSelectionMode getMode() const { return( gtk_tree_selection_get_mode( gtk_tree_view_get_selection( GTK_TREE_VIEW(list) ) ) ); }
    void setMode( GtkSelectionMode mode ) { gtk_tree_selection_set_mode( gtk_tree_view_get_selection( GTK_TREE_VIEW(list) ), mode ); }

    // ̾ѹ
    void setTitle( string newName );
  };

  /*****************************************************************
    TwinGtkList : GtkSingleList¤٤ǡѥꥹȥܥå

    "left.*" ... ¦GtkSingleList
    "right.*" ... ¦GtkSingleList
    "add.*" ... GtkEntryWithButton
    "select" ... ƥܥ
    "deselect" ... ƥܥ
    "allselect" ... ƥܥ
    "alldeselect" ... ƥܥ
  *****************************************************************/
  class TwinGtkList : public GtkBaseWidget
  {
    map<string,GtkWidget*> widgets;    // Ͽmap
    map<string,GtkSingleList*> lists;  // ꥹϿmap
    GtkEntryWithButton* entry;      // ƥȥܥåդܥ
    GtkWidget* mainBox;          // ƥʥå
    set<string> addedItem;        // ľɲä줿
    set<string> removedItem;      // ľ˺줿

    // åȤκ
    void createWidget();

    // Хåؿ
    static void event_ltr( GtkButton *button, gpointer gp );
    static void event_rtl( GtkButton *button, gpointer gp );
    static void event_all_ltr( GtkButton *button, gpointer gp );
    static void event_all_rtl( GtkButton *button, gpointer gp );
    static void event_addItem( GtkButton *button, gpointer gp );

  public:

    // 󥹥ȥ饯
    TwinGtkList( const string& dataName, GtkSelectionMode left_mode, GtkSelectionMode right_mode );
    TwinGtkList( const string& dataName, GtkSelectionMode left_mode, GtkSelectionMode right_mode, GtkSingleList::Less* _lessFunc );

    // ǥȥ饯
    virtual ~TwinGtkList();

    // ƥɲ
    unsigned int addItem( const set<string>& setS );
    bool addItem( const string& item );

    // ƥΰư
    unsigned int moveItem( const set<string>& setS );
    bool moveItem( const string& item );

    // ƥ
    void clearItem();

    // 򥢥ƥࡦƥμ
    set<string> getSelectedItem() { return( ( lists["right"] )->getSelectedItem() ); }
    set<string> getAllItem() { return( ( lists["right"] )->getAllItem() ); }

    // ľ˱/ذưƥ֤
    const set<string>& getAddItem() { return( addedItem ); }
    const set<string>& getRemovedItem() { return( removedItem ); }

    // ⡼ɤѹ
    void setMode( const string& list, GtkSelectionMode mode );
  };
};

#endif
