#ifndef PIXBUF_H
#define PIXBUF_H

#include <stdio.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <png.h>

#include <exception>
#include <iostream>

#include "gtk++.h"

namespace GtkPP
{
  using namespace std;

  template<class T> struct Coord;
  class RGB;
  class PixBuf;
  class DrawingArea;

  /*****************************************************
   㳰ѥ饹
  *****************************************************/
  class PixBufException : public exception // PixBuf㳰饹
  {
    string message;

  public:

    PixBufException( string s ) throw() : message( s ) {}
    ~PixBufException() throw() {}
    const char* what() const throw() { return( message.c_str() ); }
  };

  /*****************************************************
   ɸѹ¤
  *****************************************************/
  template<class T> struct Coord
  {
    T x;
    T y;

    // 󥹥ȥ饯
    Coord( T _x = 0, T _y = 0 ) : x( _x ), y ( _y ) {}

    // 黻Ҥ¿
    Coord& operator+=( const Coord& c ) { x += c.x; y += c.y; return( *this ); }
    Coord& operator-=( const Coord& c ) { x -= c.x; y -= c.y; return( *this ); }
    bool operator==( const Coord& c ) const { return( x == c.x && y == c.y ); }
    bool operator!=( const Coord& c ) const { return( x != c.x || y != c.y ); }

    double distance( const Coord& c ) const
    {
      return( ( sqrt( ::pow( (double)( c.x - x ), 2 ) + ::pow( (double)( c.y - y ), 2 ) ) ) );
    }
  };

  /*****************************************************
   ѥåѥ饹
  *****************************************************/
  class Color
  {
  public:

    virtual unsigned int getColor() const = 0;   // (int32)μ
    virtual void setColor( unsigned int c ) = 0; // (int32)ʬʬ򤷤ƥå

    virtual ~Color() {}
  };
  class YUV;

  class RGB : public Color
  {
    enum color { RED, GREEN, BLUE, ALPHA, SIZE }; // 
    int col[SIZE]; // ѥåȥɤʬ

    void checkCode(); // ɤΥå

    // ɤʬstaticؿ
    static unsigned char getRed( unsigned int color ) { return( ( color >> 16 ) & 0xFF ); }
    static unsigned char getGreen( unsigned int color ) { return( ( color >> 8 ) & 0xFF ); }
    static unsigned char getBlue( unsigned int color ) { return( color & 0xFF ); }

  public:

    // 󥹥ȥ饯
    RGB( unsigned char r, unsigned char g, unsigned char b, unsigned char a = 0xFF ) // RGBʬϤ
      { set( r, g, b, a ); }
    RGB( unsigned int c = 0 ) { setColor( c ); } // (int32)ι

    // 黻Ҥ¿
    RGB& operator+=( const RGB& p );
    RGB& operator-=( const RGB& p );
    RGB& operator^=( const RGB& p );
    RGB operator~();
    bool operator==( const RGB& p )
      { return( r() == p.r() && g() == p.g() && b() == p.b() ); }
    bool operator!=( const RGB& p )
      { return( r() != p.r() || g() != p.g() || b() != p.b() ); }

    // ʬμ
    void get( unsigned char& r, unsigned char& g, unsigned char& b, unsigned char& a ) const; // RGB + alpha ʬ
    void get( unsigned char& r, unsigned char& g, unsigned char& b ) const; // RGBʬ

    // ʬΥå
    void set( unsigned char r, unsigned char g, unsigned char b, unsigned char a = 0xFF );

    // ؤμ(ɤν񤭹)
    void pset( unsigned char* pixel, int n_channels ) const;
    // Ǥ򥻥å(ɤμ)
    void point( const unsigned char* pixel, int n_channels );

    // ɤμå
    unsigned int getColor() const // (int32)μ
      { return( ( r() << 16 ) | ( g() << 8 ) | b() ); }
    void setColor( unsigned int c ) // (int32)ʬʬ򤷤ƥå
      { set( getRed( c ), getGreen( c ), getBlue( c ) ); }

    // YUVѴ
    YUV toYUV() const;
    void getYUV( double& y, double& u, double& v ) const; // RGBʬYUVʬѴƼ

    // ʬ֤
    unsigned char r() const { return( col[RED] ); }
    unsigned char g() const { return( col[GREEN] ); }
    unsigned char b() const { return( col[BLUE] ); }
    unsigned char alpha() const { return( col[ALPHA] ); }
  };

  class YUV : public Color
  {
    enum color { Y, U, V, SIZE }; // 
    double col[SIZE]; // ѥåȥɤʬ
    int Alpha;        // եͥ

  public:

    YUV( double y, double u, double v, unsigned char a = 0xFF ) // YUVʬϤ
      { set( y, u, v, a ); }
    YUV( unsigned int c = 0 ) { setColor( c ); } // (int32)ι

    // ʬΥå
    void set( double y, double u, double v, unsigned char a = 0xFF );

    // RGBѴ
    RGB toRGB() const;
    void getRGB( unsigned char& r, unsigned char& g, unsigned char& b ) const; // YUVʬRGBʬѴƼ

    unsigned int getColor() const { return( toRGB().getColor() ); } // (int32)μ
    void setColor( unsigned int c ) { *this = RGB( c ).toYUV(); }    // (int32)ʬʬ򤷤ƥå

    // ʬ֤
    double y() const { return( col[Y] ); }
    double u() const { return( col[U] ); }
    double v() const { return( col[V] ); }
    unsigned char alpha() const { return( Alpha ); }
  };

  /*****************************************************
   PixBuf饹
  *****************************************************/
  class PixBuf
  {
    GdkPixbuf* pixbuf; // GDK_PIXBUF֥
    int g_width; // 
    int g_height; // ι⤵
    int g_bps; // bits_per_sample
    int g_rowstride; // rowstride
    int g_xmod; // Ԥ̤ѤΥԥ
    int g_n_channels; // n_channels
    guchar* origin; // θ򼨤ݥ

    void copy_pixbuf( const PixBuf& dest ); // PixBuf򥳥ԡ
    void input_info(); // PixBufξѿϿ

  public:

    // 󥹥ȥ饯
    PixBuf() : pixbuf( 0 ) {} // ֥Ȥۤ˽
    PixBuf( const Coord<unsigned int>& size, bool has_alpha = false ); // 礭Τߤꤷƹ
    PixBuf( string file_name ); // ե뤫鹽
    PixBuf( const char* file_name ); // ե뤫鹽
    PixBuf( const PixBuf& dest ); // ԡ󥹥ȥ饯
    PixBuf( GtkWidget* widget, const char* stock_id, GtkIconSize size, const char* detail = 0 ); // STOCK_IDλ

    // ǥȥ饯
    ~PixBuf() { g_object_unref( pixbuf ); }

    // 黻Ҥ¿
    PixBuf& operator=( const PixBuf& dest ); // 黻

    bool operator==( const PixBuf& dest );   // ˰פƤ뤫

    bool point( const Coord<int>& c, RGB& palet ) const; // ꤷɸΥѥåȤ
    bool pset( const Coord<int>& c, const RGB& palet ); // ꤷɸ˥ɥåȤ褹

    /*
      DrawingArea褹
    */
    // DrawingAreaGdkDrawableGraphics Contexts踵θɸ
    void draw( GdkDrawable* drawable, GdkGC* gc, const Coord<int>& src, const Coord<int>& dest, const Coord<unsigned int>& size ) const;
    // Graphics ContextsDrawingAreaForegroundGC˸
    void draw( GtkWidget* drawing_area, const Coord<int>& src, const Coord<int>& dest, const Coord<unsigned int>& size ) const;
    // 踵θƱ
    void draw( GtkWidget* drawing_area, const Coord<int>& c, const Coord<unsigned int>& size ) const;
    // PixBufΤ
    void draw( GtkWidget* drawing_area, const Coord<int>& c ) const;
    // 踶(0,0)˸
    void draw( GtkWidget* drawing_area ) const;

    /*
      DrawingArea礭碌褹
    */
    // DrawingAreaGdkDrawableGraphics Contexts
    void scaleDraw( GdkDrawable* drawable, const Coord<unsigned int>& size, GdkGC* gc ) const;
    // Graphics ContextsDrawingAreaForegroundGC˸
    void scaleDraw( GtkWidget* drawing_area, const Coord<unsigned int>& size ) const;

    // ƥѥ᡼ͤ
    int width() const { return( ( pixbuf != 0 ) ? g_width : 0 ); }
    int height() const { return( ( pixbuf != 0 ) ? g_height : 0 ); }
    int bps() const { return( ( pixbuf != 0 ) ? g_bps : 0 ); }
    int rowstride() const { return( ( pixbuf != 0 ) ? g_rowstride : 0 ); }
    int xmod() const { return( ( pixbuf != 0 ) ? g_xmod : 0 ); }
    int n_channels() const { return( ( pixbuf != 0 ) ? g_n_channels : 0 ); }
    bool has_alpha() const { return( ( pixbuf != 0 ) ? ( gdk_pixbuf_get_has_alpha( pixbuf ) == TRUE ) : false ); }
  };

  /*****************************************************
   DrawingArea饹
  *****************************************************/
  class DrawingArea : public GtkBaseWidget
  {
    PixBuf* pixbuf; // PixBuf֥
    GdkPixmap* pixmap; // GdkPixmap֥
    Coord<int> src; // pixbuf賫ϰ
    Coord<int> dest; // widget()賫ϰ
    Coord<unsigned int> size; // 褹Υ

    // ɸκƷ׻
    void recalc_coord();
    void recalc_coord_sub( unsigned int s_size, unsigned int d_size, int& s_coord, int& d_coord, unsigned int& size );

    // Хåؿ
    static gint configure_event( GtkWidget *widget, GdkEventExpose *event, gpointer draw ); // configure_event
    static gint expose_event( GtkWidget *widget, GdkEventExpose *event, gpointer draw ); // expose_event

    // ٥ȥϥɥΥå
    void set_event();

  public:

    // 󥹥ȥ饯
    DrawingArea( const Coord<unsigned int>& size ); // ꤷƹ
    DrawingArea( const char* file_name ); // ե̾ꤷƹ
    DrawingArea( const string& file_name );
    DrawingArea( const DrawingArea& d ); // ԡ󥹥ȥ饯
    DrawingArea( const PixBuf& p ); // PixBufꤷƹۤ

    // ǥȥ饯
    ~DrawingArea();

    // 黻Ҥ¿
    DrawingArea& operator=( const DrawingArea& d ); // 黻

    // ꤷɸPixBufФкɸ
    Coord<int> pixbufCoord( const Coord<int>& drawC ) const;

    // PixBuf֥Ȥβ
    int pixbufWidth() const { return( pixbuf->width() ); }
    int pixbufHeight() const { return( pixbuf->height() ); }

    // ꤷɸΥѥåȤ
    bool point( Coord<int> c, RGB& palet ) const { return( pixbuf->point( c, palet ) ); }
    // ꤷɸ˥ɥåȤ褹
    bool pset( Coord<int> c, const RGB& palet ) { return( pixbuf->pset( c, palet ) ); }

    // Υե
    void luShift( const Coord<int>& c ) { src += c; recalc_coord(); }
    void rdShift( const Coord<int>& c ) { src -= c; recalc_coord(); }

    // Υեå
    void reflesh( const Coord<int>& sc, const Coord<int>& dc, const Coord<unsigned int>& sz );
    void reflesh();

    // եɤ߹
    void load_file( const string& file_name );
    void load_file( const char* file_name );
  };

  /*****************************************************
   DrawingArea饹MotionNotifyEvent

     ޥɥåˤäơΰưԤ
  *****************************************************/
  class DrawingArea_MotionNotifyEvent
  {
    Coord<int> mouse;         // ˥ޥɥåΥ꡼
    DrawingArea* drawingArea; // оݤDrawingArea

    // ޥɥåˤäơΰưԤ
    static gint event( GtkWidget *widget, GdkEventMotion *event, gpointer gp );

  public:

    // 󥹥ȥ饯
    DrawingArea_MotionNotifyEvent()
      : mouse( Coord<int>( -1, -1 ) ), drawingArea( 0 ) {}

    // ٥Ȥ³Ԥ
    void connect( DrawingArea& da );
  };
}

/* ؿ */
int show_image( char*, GtkPP::DrawingArea*, char* );
int png_save( GtkPP::DrawingArea*, char* );

#endif


